@@ -1,6 +1,7 @@
 # diff_tree.py -- Utilities for diffing files and trees.
 # Copyright (C) 2010 Google, Inc.
 #
+# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
 # Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
 # General Public License as public by the Free Software Foundation; version 2.0
 # or (at your option) any later version. You can redistribute it and/or
@@ -20,20 +21,15 @@
 
 """Utilities for diffing files and trees."""
 
-from collections import (
-    defaultdict,
-    namedtuple,
-)
-
+import stat
+from collections import defaultdict, namedtuple
+from collections.abc import Iterator
 from io import BytesIO
 from itertools import chain
-import stat
-
-from dulwich.objects import (
-    S_ISGITLINK,
-    TreeEntry,
-)
+from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar
 
+from .object_store import BaseObjectStore
+from .objects import S_ISGITLINK, ObjectID, ShaFile, Tree, TreeEntry
 
 # TreeChange type constants.
 CHANGE_ADD = "add"
@@ -57,16 +53,16 @@ class TreeChange(namedtuple("TreeChange", ["type", "old", "new"])):
     """Named tuple a single change between two trees."""
 
     @classmethod
-    def add(cls, new):
+    def add(cls, new: TreeEntry) -> "TreeChange":
         return cls(CHANGE_ADD, _NULL_ENTRY, new)
 
     @classmethod
-    def delete(cls, old):
+    def delete(cls, old: TreeEntry) -> "TreeChange":
         return cls(CHANGE_DELETE, old, _NULL_ENTRY)
 
 
-def _tree_entries(path, tree):
-    result = []
+def _tree_entries(path: bytes, tree: Tree) -> list[TreeEntry]:
+    result: list[TreeEntry] = []
     if not tree:
         return result
     for entry in tree.iteritems(name_order=True):
@@ -74,13 +70,16 @@ def _tree_entries(path, tree):
     return result
 
 
-def _merge_entries(path, tree1, tree2):
+def _merge_entries(
+    path: bytes, tree1: Tree, tree2: Tree
+) -> list[tuple[TreeEntry, TreeEntry]]:
     """Merge the entries of two trees.
 
     Args:
       path: A path to prepend to all tree entry names.
       tree1: The first Tree object to iterate, or None.
       tree2: The second Tree object to iterate, or None.
+
     Returns:
       A list of pairs of TreeEntry objects for each pair of entries in
         the trees. If an entry exists in one tree but not the other, the other
@@ -114,14 +113,19 @@ def _merge_entries(path, tree1, tree2):
     return result
 
 
-def _is_tree(entry):
+def _is_tree(entry: TreeEntry) -> bool:
     mode = entry.mode
     if mode is None:
         return False
     return stat.S_ISDIR(mode)
 
 
-def walk_trees(store, tree1_id, tree2_id, prune_identical=False):
+def walk_trees(
+    store: BaseObjectStore,
+    tree1_id: Optional[ObjectID],
+    tree2_id: Optional[ObjectID],
+    prune_identical: bool = False,
+) -> Iterator[tuple[TreeEntry, TreeEntry]]:
     """Recursively walk all the entries of two trees.
 
     Iteration is depth-first pre-order, as in e.g. os.walk.
@@ -130,7 +134,8 @@ def walk_trees(store, tree1_id, tree2_id, prune_identical=False):
       store: An ObjectStore for looking up objects.
       tree1_id: The SHA of the first Tree object to iterate, or None.
       tree2_id: The SHA of the second Tree object to iterate, or None.
-      param prune_identical: If True, identical subtrees will not be walked.
+      prune_identical: If True, identical subtrees will not be walked.
+
     Returns:
       Iterator over Pairs of TreeEntry objects for each pair of entries
         in the trees and their subtrees recursively. If an entry exists in one
@@ -140,8 +145,8 @@ def walk_trees(store, tree1_id, tree2_id, prune_identical=False):
     """
     # This could be fairly easily generalized to >2 trees if we find a use
     # case.
-    mode1 = tree1_id and stat.S_IFDIR or None
-    mode2 = tree2_id and stat.S_IFDIR or None
+    mode1 = (tree1_id and stat.S_IFDIR) or None
+    mode2 = (tree2_id and stat.S_IFDIR) or None
     todo = [(TreeEntry(b"", mode1, tree1_id), TreeEntry(b"", mode2, tree2_id))]
     while todo:
         entry1, entry2 = todo.pop()
@@ -150,28 +155,41 @@ def walk_trees(store, tree1_id, tree2_id, prune_identical=False):
         if prune_identical and is_tree1 and is_tree2 and entry1 == entry2:
             continue
 
-        tree1 = is_tree1 and store[entry1.sha] or None
-        tree2 = is_tree2 and store[entry2.sha] or None
+        tree1 = (is_tree1 and store[entry1.sha]) or None
+        tree2 = (is_tree2 and store[entry2.sha]) or None
         path = entry1.path or entry2.path
-        todo.extend(reversed(_merge_entries(path, tree1, tree2)))
+
+        # Ensure trees are Tree objects before merging
+        if tree1 is not None and not isinstance(tree1, Tree):
+            tree1 = None
+        if tree2 is not None and not isinstance(tree2, Tree):
+            tree2 = None
+
+        if tree1 is not None or tree2 is not None:
+            # Use empty trees for None values
+            if tree1 is None:
+                tree1 = Tree()
+            if tree2 is None:
+                tree2 = Tree()
+            todo.extend(reversed(_merge_entries(path, tree1, tree2)))
         yield entry1, entry2
 
 
-def _skip_tree(entry, include_trees):
+def _skip_tree(entry: TreeEntry, include_trees: bool) -> TreeEntry:
     if entry.mode is None or (not include_trees and stat.S_ISDIR(entry.mode)):
         return _NULL_ENTRY
     return entry
 
 
 def tree_changes(
-    store,
-    tree1_id,
-    tree2_id,
-    want_unchanged=False,
-    rename_detector=None,
-    include_trees=False,
-    change_type_same=False,
-):
+    store: BaseObjectStore,
+    tree1_id: Optional[ObjectID],
+    tree2_id: Optional[ObjectID],
+    want_unchanged: bool = False,
+    rename_detector: Optional["RenameDetector"] = None,
+    include_trees: bool = False,
+    change_type_same: bool = False,
+) -> Iterator[TreeChange]:
     """Find the differences between the contents of two trees.
 
     Args:
@@ -184,18 +202,18 @@ def tree_changes(
       rename_detector: RenameDetector object for detecting renames.
       change_type_same: Whether to report change types in the same
         entry or as delete+add.
+
     Returns:
       Iterator over TreeChange instances for each change between the
         source and target tree.
     """
     if rename_detector is not None and tree1_id is not None and tree2_id is not None:
-        for change in rename_detector.changes_with_renames(
+        yield from rename_detector.changes_with_renames(
             tree1_id,
             tree2_id,
             want_unchanged=want_unchanged,
             include_trees=include_trees,
-        ):
-            yield change
+        )
         return
 
     entries = walk_trees(
@@ -232,18 +250,27 @@ def tree_changes(
         yield TreeChange(change_type, entry1, entry2)
 
 
-def _all_eq(seq, key, value):
+T = TypeVar("T")
+U = TypeVar("U")
+
+
+def _all_eq(seq: list[T], key: Callable[[T], U], value: U) -> bool:
     for e in seq:
         if key(e) != value:
             return False
     return True
 
 
-def _all_same(seq, key):
+def _all_same(seq: list[Any], key: Callable[[Any], Any]) -> bool:
     return _all_eq(seq[1:], key, key(seq[0]))
 
 
-def tree_changes_for_merge(store, parent_tree_ids, tree_id, rename_detector=None):
+def tree_changes_for_merge(
+    store: BaseObjectStore,
+    parent_tree_ids: list[ObjectID],
+    tree_id: ObjectID,
+    rename_detector: Optional["RenameDetector"] = None,
+) -> Iterator[list[Optional[TreeChange]]]:
     """Get the tree changes for a merge tree relative to all its parents.
 
     Args:
@@ -269,7 +296,9 @@ def tree_changes_for_merge(store, parent_tree_ids, tree_id, rename_detector=None
         for t in parent_tree_ids
     ]
     num_parents = len(parent_tree_ids)
-    changes_by_path = defaultdict(lambda: [None] * num_parents)
+    changes_by_path: dict[str, list[Optional[TreeChange]]] = defaultdict(
+        lambda: [None] * num_parents
+    )
 
     # Organize by path.
     for i, parent_changes in enumerate(all_parent_changes):
@@ -280,10 +309,10 @@ def tree_changes_for_merge(store, parent_tree_ids, tree_id, rename_detector=None
                 path = change.new.path
             changes_by_path[path][i] = change
 
-    def old_sha(c):
+    def old_sha(c: TreeChange) -> Optional[ObjectID]:
         return c.old.sha
 
-    def change_type(c):
+    def change_type(c: TreeChange) -> str:
         return c.type
 
     # Yield only conflicting changes.
@@ -305,17 +334,18 @@ def tree_changes_for_merge(store, parent_tree_ids, tree_id, rename_detector=None
 _BLOCK_SIZE = 64
 
 
-def _count_blocks(obj):
+def _count_blocks(obj: ShaFile) -> dict[int, int]:
     """Count the blocks in an object.
 
     Splits the data into blocks either on lines or <=64-byte chunks of lines.
 
     Args:
       obj: The object to count blocks for.
+
     Returns:
       A dict of block hashcode -> total bytes occurring.
     """
-    block_counts = defaultdict(int)
+    block_counts: dict[int, int] = defaultdict(int)
     block = BytesIO()
     n = 0
 
@@ -326,10 +356,10 @@ def _count_blocks(obj):
     block_getvalue = block.getvalue
 
     for c in chain.from_iterable(obj.as_raw_chunks()):
-        c = c.to_bytes(1, "big")
-        block_write(c)
+        cb = c.to_bytes(1, "big")
+        block_write(cb)
         n += 1
-        if c == b"\n" or n == _BLOCK_SIZE:
+        if cb == b"\n" or n == _BLOCK_SIZE:
             value = block_getvalue()
             block_counts[hash(value)] += len(value)
             block_seek(0)
@@ -341,12 +371,13 @@ def _count_blocks(obj):
     return block_counts
 
 
-def _common_bytes(blocks1, blocks2):
+def _common_bytes(blocks1: dict[int, int], blocks2: dict[int, int]) -> int:
     """Count the number of common bytes in two block count dicts.
 
     Args:
-      block1: The first dict of block hashcode -> total bytes.
-      block2: The second dict of block hashcode -> total bytes.
+      blocks1: The first dict of block hashcode -> total bytes.
+      blocks2: The second dict of block hashcode -> total bytes.
+
     Returns:
       The number of bytes in common between blocks1 and blocks2. This is
       only approximate due to possible hash collisions.
@@ -362,7 +393,11 @@ def _common_bytes(blocks1, blocks2):
     return score
 
 
-def _similarity_score(obj1, obj2, block_cache=None):
+def _similarity_score(
+    obj1: ShaFile,
+    obj2: ShaFile,
+    block_cache: Optional[dict[ObjectID, dict[int, int]]] = None,
+) -> int:
     """Compute a similarity score for two objects.
 
     Args:
@@ -370,6 +405,7 @@ def _similarity_score(obj1, obj2, block_cache=None):
       obj2: The second object to score.
       block_cache: An optional dict of SHA to block counts to cache
         results between calls.
+
     Returns:
       The similarity score between the two objects, defined as the
         number of bytes in common between the two objects divided by the
@@ -389,7 +425,7 @@ def _similarity_score(obj1, obj2, block_cache=None):
     return int(float(common_bytes) * _MAX_SCORE / max_size)
 
 
-def _tree_change_key(entry):
+def _tree_change_key(entry: TreeChange) -> tuple[bytes, bytes]:
     # Sort by old path then new path. If only one exists, use it for both keys.
     path1 = entry.old.path
     path2 = entry.new.path
@@ -400,17 +436,22 @@ def _tree_change_key(entry):
     return (path1, path2)
 
 
-class RenameDetector(object):
+class RenameDetector:
     """Object for handling rename detection between two trees."""
 
+    _adds: list[TreeChange]
+    _deletes: list[TreeChange]
+    _changes: list[TreeChange]
+    _candidates: list[tuple[int, TreeChange]]
+
     def __init__(
         self,
-        store,
-        rename_threshold=RENAME_THRESHOLD,
-        max_files=MAX_FILES,
-        rewrite_threshold=REWRITE_THRESHOLD,
-        find_copies_harder=False,
-    ):
+        store: BaseObjectStore,
+        rename_threshold: int = RENAME_THRESHOLD,
+        max_files: Optional[int] = MAX_FILES,
+        rewrite_threshold: Optional[int] = REWRITE_THRESHOLD,
+        find_copies_harder: bool = False,
+    ) -> None:
         """Initialize the rename detector.
 
         Args:
@@ -435,12 +476,12 @@ class RenameDetector(object):
         self._find_copies_harder = find_copies_harder
         self._want_unchanged = False
 
-    def _reset(self):
+    def _reset(self) -> None:
         self._adds = []
         self._deletes = []
         self._changes = []
 
-    def _should_split(self, change):
+    def _should_split(self, change: TreeChange) -> bool:
         if (
             self._rewrite_threshold is None
             or change.type != CHANGE_MODIFY
@@ -451,7 +492,7 @@ class RenameDetector(object):
         new_obj = self._store[change.new.sha]
         return _similarity_score(old_obj, new_obj) < self._rewrite_threshold
 
-    def _add_change(self, change):
+    def _add_change(self, change: TreeChange) -> None:
         if change.type == CHANGE_ADD:
             self._adds.append(change)
         elif change.type == CHANGE_DELETE:
@@ -470,7 +511,9 @@ class RenameDetector(object):
         else:
             self._changes.append(change)
 
-    def _collect_changes(self, tree1_id, tree2_id):
+    def _collect_changes(
+        self, tree1_id: Optional[ObjectID], tree2_id: Optional[ObjectID]
+    ) -> None:
         want_unchanged = self._find_copies_harder or self._want_unchanged
         for change in tree_changes(
             self._store,
@@ -481,11 +524,11 @@ class RenameDetector(object):
         ):
             self._add_change(change)
 
-    def _prune(self, add_paths, delete_paths):
+    def _prune(self, add_paths: set[bytes], delete_paths: set[bytes]) -> None:
         self._adds = [a for a in self._adds if a.new.path not in add_paths]
         self._deletes = [d for d in self._deletes if d.old.path not in delete_paths]
 
-    def _find_exact_renames(self):
+    def _find_exact_renames(self) -> None:
         add_map = defaultdict(list)
         for add in self._adds:
             add_map[add.new.sha].append(add.new)
@@ -506,7 +549,7 @@ class RenameDetector(object):
                 if is_delete:
                     delete_paths.add(old.path)
                 add_paths.add(new.path)
-                new_type = is_delete and CHANGE_RENAME or CHANGE_COPY
+                new_type = (is_delete and CHANGE_RENAME) or CHANGE_COPY
                 self._changes.append(TreeChange(new_type, old, new))
 
             num_extra_adds = len(sha_adds) - len(sha_deletes)
@@ -518,10 +561,14 @@ class RenameDetector(object):
                     self._changes.append(TreeChange(CHANGE_COPY, old, new))
         self._prune(add_paths, delete_paths)
 
-    def _should_find_content_renames(self):
-        return len(self._adds) * len(self._deletes) <= self._max_files ** 2
+    def _should_find_content_renames(self) -> bool:
+        if self._max_files is None:
+            return True
+        return len(self._adds) * len(self._deletes) <= self._max_files**2
 
-    def _rename_type(self, check_paths, delete, add):
+    def _rename_type(
+        self, check_paths: bool, delete: TreeChange, add: TreeChange
+    ) -> str:
         if check_paths and delete.old.path == add.new.path:
             # If the paths match, this must be a split modify, so make sure it
             # comes out as a modify.
@@ -533,7 +580,7 @@ class RenameDetector(object):
             return CHANGE_COPY
         return CHANGE_RENAME
 
-    def _find_content_rename_candidates(self):
+    def _find_content_rename_candidates(self) -> None:
         candidates = self._candidates = []
         # TODO: Optimizations:
         #  - Compare object sizes before counting blocks.
@@ -562,7 +609,7 @@ class RenameDetector(object):
                     rename = TreeChange(new_type, delete.old, add.new)
                     candidates.append((-score, rename))
 
-    def _choose_content_renames(self):
+    def _choose_content_renames(self) -> None:
         # Sort scores from highest to lowest, but keep names in ascending
         # order.
         self._candidates.sort()
@@ -586,7 +633,7 @@ class RenameDetector(object):
             self._changes.append(change)
         self._prune(add_paths, delete_paths)
 
-    def _join_modifies(self):
+    def _join_modifies(self) -> None:
         if self._rewrite_threshold is None:
             return
 
@@ -604,7 +651,7 @@ class RenameDetector(object):
         self._deletes = [a for a in self._deletes if a.new.path not in modifies]
         self._changes += modifies.values()
 
-    def _sorted_changes(self):
+    def _sorted_changes(self) -> list[TreeChange]:
         result = []
         result.extend(self._adds)
         result.extend(self._deletes)
@@ -612,14 +659,18 @@ class RenameDetector(object):
         result.sort(key=_tree_change_key)
         return result
 
-    def _prune_unchanged(self):
+    def _prune_unchanged(self) -> None:
         if self._want_unchanged:
             return
         self._deletes = [d for d in self._deletes if d.type != CHANGE_UNCHANGED]
 
     def changes_with_renames(
-        self, tree1_id, tree2_id, want_unchanged=False, include_trees=False
-    ):
+        self,
+        tree1_id: Optional[ObjectID],
+        tree2_id: Optional[ObjectID],
+        want_unchanged: bool = False,
+        include_trees: bool = False,
+    ) -> list[TreeChange]:
         """Iterate TreeChanges between two tree SHAs, with rename detection."""
         self._reset()
         self._want_unchanged = want_unchanged
@@ -637,12 +688,27 @@ class RenameDetector(object):
 _is_tree_py = _is_tree
 _merge_entries_py = _merge_entries
 _count_blocks_py = _count_blocks
-try:
-    # Try to import C versions
-    from dulwich._diff_tree import (  # type: ignore
-        _is_tree,
-        _merge_entries,
-        _count_blocks,
-    )
-except ImportError:
+
+if TYPE_CHECKING:
+    # For type checking, use the Python implementations
     pass
+else:
+    # At runtime, try to import Rust extensions
+    try:
+        # Try to import Rust versions
+        from dulwich._diff_tree import (
+            _count_blocks as _rust_count_blocks,
+        )
+        from dulwich._diff_tree import (
+            _is_tree as _rust_is_tree,
+        )
+        from dulwich._diff_tree import (
+            _merge_entries as _rust_merge_entries,
+        )
+
+        # Override with Rust versions
+        _count_blocks = _rust_count_blocks
+        _is_tree = _rust_is_tree
+        _merge_entries = _rust_merge_entries
+    except ImportError:
+        pass
