@@ -1,6 +1,7 @@
 # line_ending.py -- Line ending conversion functions
 # Copyright (C) 2018-2018 Boris Feld <boris.feld@comet.ml>
 #
+# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
 # Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
 # General Public License as public by the Free Software Foundation; version 2.0
 # or (at your option) any later version. You can redistribute it and/or
@@ -17,7 +18,7 @@
 # and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
 # License, Version 2.0.
 #
-"""All line-ending related functions, from conversions to config processing
+r"""All line-ending related functions, from conversions to config processing.
 
 Line-ending normalization is a complex beast. Here is some notes and details
 about how it seems to work.
@@ -61,23 +62,23 @@ There is multiple variables that impact the normalization.
 First, a repository can contains a ``.gitattributes`` file (or more than one...)
 that can further customize the operation on some file patterns, for example:
 
-    \\*.txt text
+    \*.txt text
 
 Force all ``.txt`` files to be treated as text files and to have their lines
 endings normalized.
 
-    \\*.jpg -text
+    \*.jpg -text
 
 Force all ``.jpg`` files to be treated as binary files and to not have their
 lines endings converted.
 
-    \\*.vcproj text eol=crlf
+    \*.vcproj text eol=crlf
 
 Force all ``.vcproj`` files to be treated as text files and to have their lines
 endings converted into ``CRLF`` in working directory no matter the native EOL of
 the platform.
 
-    \\*.sh text eol=lf
+    \*.sh text eol=lf
 
 Force all ``.sh`` files to be treated as text files and to have their lines
 endings converted into ``LF`` in working directory no matter the native EOL of
@@ -86,7 +87,7 @@ the platform.
 If the ``eol`` attribute is not defined, Git uses the ``core.eol`` configuration
 value described later.
 
-    \\* text=auto
+    \* text=auto
 
 Force all files to be scanned by the text file heuristic detection and to have
 their line endings normalized in case they are detected as text files.
@@ -136,16 +137,22 @@ Sources:
 - https://adaptivepatchwork.com/2012/03/01/mind-the-end-of-your-line/
 """
 
-from dulwich.object_store import iter_tree_contents
-from dulwich.objects import Blob
-from dulwich.patch import is_binary
+from typing import TYPE_CHECKING, Any, Callable, Optional, Union
+
+if TYPE_CHECKING:
+    from .config import StackedConfig
+    from .object_store import BaseObjectStore
+
+from .object_store import iter_tree_contents
+from .objects import Blob, ObjectID
+from .patch import is_binary
 
 CRLF = b"\r\n"
 LF = b"\n"
 
 
-def convert_crlf_to_lf(text_hunk):
-    """Convert CRLF in text hunk into LF
+def convert_crlf_to_lf(text_hunk: bytes) -> bytes:
+    """Convert CRLF in text hunk into LF.
 
     Args:
       text_hunk: A bytes string representing a text hunk
@@ -154,36 +161,66 @@ def convert_crlf_to_lf(text_hunk):
     return text_hunk.replace(CRLF, LF)
 
 
-def convert_lf_to_crlf(text_hunk):
-    """Convert LF in text hunk into CRLF
+def convert_lf_to_crlf(text_hunk: bytes) -> bytes:
+    """Convert LF in text hunk into CRLF.
 
     Args:
       text_hunk: A bytes string representing a text hunk
     Returns: The text hunk with the same type, with LF replaced into CRLF
     """
-    # TODO find a more efficient way of doing it
-    intermediary = text_hunk.replace(CRLF, LF)
-    return intermediary.replace(LF, CRLF)
+    # Single-pass conversion: split on LF and join with CRLF
+    # This avoids the double replacement issue
+    parts = text_hunk.split(LF)
+    # Remove any trailing CR to avoid CRCRLF
+    cleaned_parts = []
+    for i, part in enumerate(parts):
+        if i < len(parts) - 1 and part.endswith(b"\r"):
+            cleaned_parts.append(part[:-1])
+        else:
+            cleaned_parts.append(part)
+    return CRLF.join(cleaned_parts)
 
 
-def get_checkout_filter(core_eol, core_autocrlf, git_attributes):
-    """Returns the correct checkout filter based on the passed arguments"""
+def get_checkout_filter(
+    core_eol: str, core_autocrlf: Union[bool, str], git_attributes: dict[str, Any]
+) -> Optional[Callable[[bytes], bytes]]:
+    """Returns the correct checkout filter based on the passed arguments."""
     # TODO this function should process the git_attributes for the path and if
     # the text attribute is not defined, fallback on the
     # get_checkout_filter_autocrlf function with the autocrlf value
-    return get_checkout_filter_autocrlf(core_autocrlf)
+    if isinstance(core_autocrlf, bool):
+        autocrlf_bytes = b"true" if core_autocrlf else b"false"
+    else:
+        autocrlf_bytes = (
+            core_autocrlf.encode("ascii")
+            if isinstance(core_autocrlf, str)
+            else core_autocrlf
+        )
+    return get_checkout_filter_autocrlf(autocrlf_bytes)
 
 
-def get_checkin_filter(core_eol, core_autocrlf, git_attributes):
-    """Returns the correct checkin filter based on the passed arguments"""
+def get_checkin_filter(
+    core_eol: str, core_autocrlf: Union[bool, str], git_attributes: dict[str, Any]
+) -> Optional[Callable[[bytes], bytes]]:
+    """Returns the correct checkin filter based on the passed arguments."""
     # TODO this function should process the git_attributes for the path and if
     # the text attribute is not defined, fallback on the
     # get_checkin_filter_autocrlf function with the autocrlf value
-    return get_checkin_filter_autocrlf(core_autocrlf)
+    if isinstance(core_autocrlf, bool):
+        autocrlf_bytes = b"true" if core_autocrlf else b"false"
+    else:
+        autocrlf_bytes = (
+            core_autocrlf.encode("ascii")
+            if isinstance(core_autocrlf, str)
+            else core_autocrlf
+        )
+    return get_checkin_filter_autocrlf(autocrlf_bytes)
 
 
-def get_checkout_filter_autocrlf(core_autocrlf):
-    """Returns the correct checkout filter base on autocrlf value
+def get_checkout_filter_autocrlf(
+    core_autocrlf: bytes,
+) -> Optional[Callable[[bytes], bytes]]:
+    """Returns the correct checkout filter base on autocrlf value.
 
     Args:
       core_autocrlf: The bytes configuration value of core.autocrlf.
@@ -191,15 +228,16 @@ def get_checkout_filter_autocrlf(core_autocrlf):
     Returns: Either None if no filter has to be applied or a function
         accepting a single argument, a binary text hunk
     """
-
     if core_autocrlf == b"true":
         return convert_lf_to_crlf
 
     return None
 
 
-def get_checkin_filter_autocrlf(core_autocrlf):
-    """Returns the correct checkin filter base on autocrlf value
+def get_checkin_filter_autocrlf(
+    core_autocrlf: bytes,
+) -> Optional[Callable[[bytes], bytes]]:
+    """Returns the correct checkin filter base on autocrlf value.
 
     Args:
       core_autocrlf: The bytes configuration value of core.autocrlf.
@@ -207,7 +245,6 @@ def get_checkin_filter_autocrlf(core_autocrlf):
     Returns: Either None if no filter has to be applied or a function
         accepting a single argument, a binary text hunk
     """
-
     if core_autocrlf == b"true" or core_autocrlf == b"input":
         return convert_crlf_to_lf
 
@@ -217,21 +254,34 @@ def get_checkin_filter_autocrlf(core_autocrlf):
 
 class BlobNormalizer:
     """An object to store computation result of which filter to apply based
-    on configuration, gitattributes, path and operation (checkin or checkout)
+    on configuration, gitattributes, path and operation (checkin or checkout).
     """
 
-    def __init__(self, config_stack, gitattributes):
+    def __init__(
+        self, config_stack: "StackedConfig", gitattributes: dict[str, Any]
+    ) -> None:
         self.config_stack = config_stack
         self.gitattributes = gitattributes
 
         # Compute which filters we needs based on parameters
         try:
-            core_eol = config_stack.get("core", "eol")
+            core_eol_raw = config_stack.get("core", "eol")
+            core_eol: str = (
+                core_eol_raw.decode("ascii")
+                if isinstance(core_eol_raw, bytes)
+                else core_eol_raw
+            )
         except KeyError:
             core_eol = "native"
 
         try:
-            core_autocrlf = config_stack.get("core", "autocrlf").lower()
+            core_autocrlf_raw = config_stack.get("core", "autocrlf")
+            if isinstance(core_autocrlf_raw, bytes):
+                core_autocrlf: Union[bool, str] = core_autocrlf_raw.decode(
+                    "ascii"
+                ).lower()
+            else:
+                core_autocrlf = core_autocrlf_raw.lower()
         except KeyError:
             core_autocrlf = False
 
@@ -242,8 +292,8 @@ class BlobNormalizer:
             core_eol, core_autocrlf, self.gitattributes
         )
 
-    def checkin_normalize(self, blob, tree_path):
-        """Normalize a blob during a checkin operation"""
+    def checkin_normalize(self, blob: Blob, tree_path: bytes) -> Blob:
+        """Normalize a blob during a checkin operation."""
         if self.fallback_write_filter is not None:
             return normalize_blob(
                 blob, self.fallback_write_filter, binary_detection=True
@@ -251,8 +301,8 @@ class BlobNormalizer:
 
         return blob
 
-    def checkout_normalize(self, blob, tree_path):
-        """Normalize a blob during a checkout operation"""
+    def checkout_normalize(self, blob: Blob, tree_path: bytes) -> Blob:
+        """Normalize a blob during a checkout operation."""
         if self.fallback_read_filter is not None:
             return normalize_blob(
                 blob, self.fallback_read_filter, binary_detection=True
@@ -261,10 +311,12 @@ class BlobNormalizer:
         return blob
 
 
-def normalize_blob(blob, conversion, binary_detection):
+def normalize_blob(
+    blob: Blob, conversion: Callable[[bytes], bytes], binary_detection: bool
+) -> Blob:
     """Takes a blob as input returns either the original blob if
     binary_detection is True and the blob content looks like binary, else
-    return a new blob with converted data
+    return a new blob with converted data.
     """
     # Read the original blob
     data = blob.data
@@ -286,17 +338,22 @@ def normalize_blob(blob, conversion, binary_detection):
 
 
 class TreeBlobNormalizer(BlobNormalizer):
-    def __init__(self, config_stack, git_attributes, object_store, tree=None):
+    def __init__(
+        self,
+        config_stack: "StackedConfig",
+        git_attributes: dict[str, Any],
+        object_store: "BaseObjectStore",
+        tree: Optional[ObjectID] = None,
+    ) -> None:
         super().__init__(config_stack, git_attributes)
         if tree:
             self.existing_paths = {
-                name
-                for name, _, _ in iter_tree_contents(object_store, tree)
+                name for name, _, _ in iter_tree_contents(object_store, tree)
             }
         else:
             self.existing_paths = set()
 
-    def checkin_normalize(self, blob, tree_path):
+    def checkin_normalize(self, blob: Blob, tree_path: bytes) -> Blob:
         # Existing files should only be normalized on checkin if it was
         # previously normalized on checkout
         if (
