@@ -2,6 +2,7 @@
 # Copyright (C) 2010 Google, Inc.
 # Copyright (C) 2012 Jelmer Vernooij <jelmer@jelmer.uk>
 #
+# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
 # Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
 # General Public License as public by the Free Software Foundation; version 2.0
 # or (at your option) any later version. You can redistribute it and/or
@@ -25,17 +26,28 @@ import os
 import re
 import sys
 import time
+from collections.abc import Iterator
 from io import BytesIO
-from typing import List, Optional, Tuple
+from typing import Callable, ClassVar, Optional
 from urllib.parse import parse_qs
-from wsgiref.simple_server import (ServerHandler, WSGIRequestHandler,
-                                   WSGIServer, make_server)
+from wsgiref.simple_server import (
+    ServerHandler,
+    WSGIRequestHandler,
+    WSGIServer,
+    make_server,
+)
 
 from dulwich import log_utils
-from dulwich.protocol import ReceivableProtocol
-from dulwich.repo import BaseRepo, NotGitRepository, Repo
-from dulwich.server import (DEFAULT_HANDLERS, DictBackend, generate_info_refs,
-                            generate_objects_info_packs)
+
+from .protocol import ReceivableProtocol
+from .repo import BaseRepo, NotGitRepository, Repo
+from .server import (
+    DEFAULT_HANDLERS,
+    Backend,
+    DictBackend,
+    generate_info_refs,
+    generate_objects_info_packs,
+)
 
 logger = log_utils.getLogger(__name__)
 
@@ -90,7 +102,7 @@ def date_time_string(timestamp: Optional[float] = None) -> str:
     if timestamp is None:
         timestamp = time.time()
     year, month, day, hh, mm, ss, wd = time.gmtime(timestamp)[:7]
-    return "%s, %02d %3s %4d %02d:%02d:%02d GMD" % (
+    return "%s, %02d %3s %4d %02d:%02d:%02d GMD" % (  # noqa: UP031
         weekdays[wd],
         day,
         months[month],
@@ -207,7 +219,7 @@ def get_info_refs(req, backend, mat):
             yield req.forbidden("Unsupported service")
             return
         req.nocache()
-        write = req.respond(HTTP_OK, "application/x-%s-advertisement" % service)
+        write = req.respond(HTTP_OK, f"application/x-{service}-advertisement")
         proto = ReceivableProtocol(BytesIO().read, write)
         handler = handler_cls(
             backend,
@@ -246,10 +258,11 @@ def _chunk_iter(f):
 
 
 class ChunkReader:
+    """Reader for chunked transfer encoding streams."""
 
-    def __init__(self, f):
+    def __init__(self, f) -> None:
         self._iter = _chunk_iter(f)
-        self._buffer = []
+        self._buffer: list[bytes] = []
 
     def read(self, n):
         while sum(map(len, self._buffer)) < n:
@@ -257,7 +270,7 @@ class ChunkReader:
                 self._buffer.append(next(self._iter))
             except StopIteration:
                 break
-        f = b''.join(self._buffer)
+        f = b"".join(self._buffer)
         ret = f[:n]
         self._buffer = [f[n:]]
         return ret
@@ -271,7 +284,7 @@ class _LengthLimitedFile:
     but not implemented in wsgiref as of 2.5.
     """
 
-    def __init__(self, input, max_bytes):
+    def __init__(self, input, max_bytes) -> None:
         self._input = input
         self._bytes_avail = max_bytes
 
@@ -299,8 +312,8 @@ def handle_service_request(req, backend, mat):
         yield req.not_found(str(e))
         return
     req.nocache()
-    write = req.respond(HTTP_OK, "application/x-%s-result" % service)
-    if req.environ.get('HTTP_TRANSFER_ENCODING') == 'chunked':
+    write = req.respond(HTTP_OK, f"application/x-{service}-result")
+    if req.environ.get("HTTP_TRANSFER_ENCODING") == "chunked":
         read = ChunkReader(req.environ["wsgi.input"]).read
     else:
         read = req.environ["wsgi.input"].read
@@ -318,15 +331,17 @@ class HTTPGitRequest:
       environ: the WSGI environment for the request.
     """
 
-    def __init__(self, environ, start_response, dumb: bool = False, handlers=None):
+    def __init__(
+        self, environ, start_response, dumb: bool = False, handlers=None
+    ) -> None:
         self.environ = environ
         self.dumb = dumb
         self.handlers = handlers
         self._start_response = start_response
-        self._cache_headers: List[Tuple[str, str]] = []
-        self._headers: List[Tuple[str, str]] = []
+        self._cache_headers: list[tuple[str, str]] = []
+        self._headers: list[tuple[str, str]] = []
 
-    def add_header(self, name, value):
+    def add_header(self, name, value) -> None:
         """Add a header to the response."""
         self._headers.append((name, value))
 
@@ -334,7 +349,7 @@ class HTTPGitRequest:
         self,
         status: str = HTTP_OK,
         content_type: Optional[str] = None,
-        headers: Optional[List[Tuple[str, str]]] = None,
+        headers: Optional[list[tuple[str, str]]] = None,
     ):
         """Begin a response with the given status and other headers."""
         if headers:
@@ -382,7 +397,12 @@ class HTTPGitApplication:
       backend: the Backend object backing this application
     """
 
-    services = {
+    services: ClassVar[
+        dict[
+            tuple[str, re.Pattern],
+            Callable[[HTTPGitRequest, Backend, re.Match], Iterator[bytes]],
+        ]
+    ] = {
         ("GET", re.compile("/HEAD$")): get_text_file,
         ("GET", re.compile("/info/refs$")): get_info_refs,
         ("GET", re.compile("/objects/info/alternates$")): get_text_file,
@@ -404,7 +424,9 @@ class HTTPGitApplication:
         ("POST", re.compile("/git-receive-pack$")): handle_service_request,
     }
 
-    def __init__(self, backend, dumb: bool = False, handlers=None, fallback_app=None):
+    def __init__(
+        self, backend, dumb: bool = False, handlers=None, fallback_app=None
+    ) -> None:
         self.backend = backend
         self.dumb = dumb
         self.handlers = dict(DEFAULT_HANDLERS)
@@ -442,11 +464,12 @@ class GunzipFilter:
     passing on to the underlying application.
     """
 
-    def __init__(self, application):
+    def __init__(self, application) -> None:
         self.app = application
 
     def __call__(self, environ, start_response):
         import gzip
+
         if environ.get("HTTP_CONTENT_ENCODING", "") == "gzip":
             environ["wsgi.input"] = gzip.GzipFile(
                 filename=None, fileobj=environ["wsgi.input"], mode="rb"
@@ -463,7 +486,7 @@ class LimitedInputFilter:
     specified in Content-Length.
     """
 
-    def __init__(self, application):
+    def __init__(self, application) -> None:
         self.app = application
 
     def __call__(self, environ, start_response):
@@ -491,58 +514,61 @@ def make_wsgi_chain(*args, **kwargs):
 class ServerHandlerLogger(ServerHandler):
     """ServerHandler that uses dulwich's logger for logging exceptions."""
 
-    def log_exception(self, exc_info):
+    def log_exception(self, exc_info) -> None:
         logger.exception(
             "Exception happened during processing of request",
             exc_info=exc_info,
         )
 
-    def log_message(self, format, *args):
+    def log_message(self, format, *args) -> None:
         logger.info(format, *args)
 
-    def log_error(self, *args):
+    def log_error(self, *args) -> None:
         logger.error(*args)
 
 
 class WSGIRequestHandlerLogger(WSGIRequestHandler):
     """WSGIRequestHandler that uses dulwich's logger for logging exceptions."""
 
-    def log_exception(self, exc_info):
+    def log_exception(self, exc_info) -> None:
         logger.exception(
             "Exception happened during processing of request",
             exc_info=exc_info,
         )
 
-    def log_message(self, format, *args):
+    def log_message(self, format, *args) -> None:
         logger.info(format, *args)
 
-    def log_error(self, *args):
+    def log_error(self, *args) -> None:
         logger.error(*args)
 
-    def handle(self):
-        """Handle a single HTTP request"""
-
+    def handle(self) -> None:
+        """Handle a single HTTP request."""
         self.raw_requestline = self.rfile.readline()
         if not self.parse_request():  # An error code has been sent, just exit
             return
 
         handler = ServerHandlerLogger(
-            self.rfile, self.wfile, self.get_stderr(), self.get_environ()
+            self.rfile,
+            self.wfile,  # type: ignore
+            self.get_stderr(),
+            self.get_environ(),
         )
-        handler.request_handler = self  # backpointer for logging
-        handler.run(self.server.get_app())
+        handler.request_handler = self  # type: ignore  # backpointer for logging
+        handler.run(self.server.get_app())  # type: ignore
 
 
 class WSGIServerLogger(WSGIServer):
-    def handle_error(self, request, client_address):
-        """Handle an error. """
+    """WSGIServer that uses dulwich's logger for error handling."""
+
+    def handle_error(self, request, client_address) -> None:
+        """Handle an error."""
         logger.exception(
-            "Exception happened during processing of request from %s"
-            % str(client_address)
+            f"Exception happened during processing of request from {client_address!s}"
         )
 
 
-def main(argv=sys.argv):
+def main(argv=sys.argv) -> None:
     """Entry point for starting an HTTP git server."""
     import optparse
 
